/* 
** NetXMS - Network Management System
** Copyright (C) 2005-2009 Victor Kirhenshtein
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
** File: nxsl_classes.h
**
**/

#ifndef _nxsl_classes_h_
#define _nxsl_classes_h_

#include <nms_threads.h>
#include <nms_util.h>
#include <nxcpapi.h>


//
// Constants
//

#define MAX_CLASS_NAME     64


//
// Data types
//

#define NXSL_DT_NULL       0
#define NXSL_DT_OBJECT     1
#define NXSL_DT_ARRAY      2
#define NXSL_DT_STRING     3
#define NXSL_DT_REAL       4
#define NXSL_DT_INT32      5
#define NXSL_DT_INT64      6
#define NXSL_DT_UINT32     7
#define NXSL_DT_UINT64     8


//
// Simple stack class
//

class LIBNXSL_EXPORTABLE NXSL_Stack
{
private:
   int m_nStackSize;
   int m_nStackPos;
   void **m_ppData;

public:
   NXSL_Stack(void);
   ~NXSL_Stack();

   void Push(void *pData);
   void *Pop(void);
   void *Peek(void);
   void **PeekList(int nLevel) { return &m_ppData[m_nStackPos - nLevel]; }

   int Size(void) { return m_nStackPos; }
};


//
// Class representing NXSL class
//

class NXSL_Value;
class NXSL_Object;

class LIBNXSL_EXPORTABLE NXSL_Class
{
protected:
   char m_szName[MAX_CLASS_NAME];

public:
   NXSL_Class();
   virtual ~NXSL_Class();

   virtual NXSL_Value *GetAttr(NXSL_Object *pObject, char *pszAttr);
   virtual BOOL SetAttr(NXSL_Object *pObject, char *pszAttr, NXSL_Value *pValue);

	virtual void OnObjectDelete(NXSL_Object *object);

   char *Name(void) { return m_szName; }
};


//
// Object instance
//

class LIBNXSL_EXPORTABLE NXSL_Object
{
private:
   NXSL_Class *m_pClass;
   void *m_pData;

public:
   NXSL_Object(NXSL_Object *pObject);
   NXSL_Object(NXSL_Class *pClass, void *pData);
   ~NXSL_Object();

   NXSL_Class *Class() { return m_pClass; }
   void *Data() { return m_pData; }
};


//
// Array
//

struct NXSL_ArrayElement
{
	int index;
	NXSL_Value *value;
};

class LIBNXSL_EXPORTABLE NXSL_Array
{
private:
	int m_refCount;
	int m_size;
	int m_allocated;
	NXSL_ArrayElement *m_data;

public:
	NXSL_Array();
	NXSL_Array(NXSL_Array *src);
	~NXSL_Array();

	void IncRefCount() { m_refCount++; }
	void DecRefCount() { m_refCount--; }
	BOOL IsUnused() { return m_refCount < 1; }

	void Set(int index, NXSL_Value *value);
	NXSL_Value *Get(int index);
};


//
// Variable or constant value
//

class LIBNXSL_EXPORTABLE NXSL_Value
{
protected:
   DWORD m_dwStrLen;
   char *m_pszValStr;
   BYTE m_nDataType;
   BYTE m_bStringIsValid;
   union
   {
      LONG nInt32;
      DWORD uInt32;
      INT64 nInt64;
      QWORD uInt64;
      double dReal;
      NXSL_Object *pObject;
		NXSL_Array *pArray;
   } m_value;

   void UpdateNumber(void);
   void UpdateString(void);

   void InvalidateString(void)
   {
      safe_free(m_pszValStr);
      m_pszValStr = NULL;
      m_bStringIsValid = FALSE;
   }

public:
   NXSL_Value(void);
   NXSL_Value(const NXSL_Value *);
   NXSL_Value(NXSL_Object *pObject);
   NXSL_Value(NXSL_Array *pArray);
   NXSL_Value(LONG nValue);
   NXSL_Value(INT64 nValue);
   NXSL_Value(DWORD uValue);
   NXSL_Value(QWORD uValue);
   NXSL_Value(double dValue);
   NXSL_Value(const TCHAR *pszValue);
   NXSL_Value(const TCHAR *pszValue, DWORD dwLen);
   ~NXSL_Value();

   void Set(LONG nValue);

   BOOL Convert(int nDataType);
   int DataType(void) { return m_nDataType; }

   BOOL IsNull(void) { return (m_nDataType == NXSL_DT_NULL); }
   BOOL IsObject(void) { return (m_nDataType == NXSL_DT_OBJECT); }
   BOOL IsArray(void) { return (m_nDataType == NXSL_DT_ARRAY); }
   BOOL IsString(void) { return (m_nDataType >= NXSL_DT_STRING); }
   BOOL IsNumeric(void) { return (m_nDataType > NXSL_DT_STRING); }
   BOOL IsReal(void) { return (m_nDataType == NXSL_DT_REAL); }
   BOOL IsInteger(void) { return (m_nDataType > NXSL_DT_REAL); }
   BOOL IsUnsigned(void) { return (m_nDataType >= NXSL_DT_UINT32); }
   BOOL IsZero(void);
   BOOL IsNonZero(void);

   char *GetValueAsString(DWORD *pdwLen);
   char *GetValueAsCString(void);
   LONG GetValueAsInt32(void);
   DWORD GetValueAsUInt32(void);
   INT64 GetValueAsInt64(void);
   QWORD GetValueAsUInt64(void);
   double GetValueAsReal(void);
   NXSL_Object *GetValueAsObject(void) { return (m_nDataType == NXSL_DT_OBJECT) ? m_value.pObject : NULL; }
   NXSL_Array *GetValueAsArray(void) { return (m_nDataType == NXSL_DT_ARRAY) ? m_value.pArray : NULL; }

   void Concatenate(char *pszString, DWORD dwLen);
   
   void Increment(void);
   void Decrement(void);
   void Negate(void);
   void BitNot(void);

   void Add(NXSL_Value *pVal);
   void Sub(NXSL_Value *pVal);
   void Mul(NXSL_Value *pVal);
   void Div(NXSL_Value *pVal);
   void Rem(NXSL_Value *pVal);
   void BitAnd(NXSL_Value *pVal);
   void BitOr(NXSL_Value *pVal);
   void BitXor(NXSL_Value *pVal);
   void LShift(int nBits);
   void RShift(int nBits);

   BOOL EQ(NXSL_Value *pVal);
   BOOL LT(NXSL_Value *pVal);
   BOOL LE(NXSL_Value *pVal);
   BOOL GT(NXSL_Value *pVal);
   BOOL GE(NXSL_Value *pVal);
};


//
// Function structure
//

struct NXSL_Function
{
   char m_szName[MAX_FUNCTION_NAME];
   DWORD m_dwAddr;
};


//
// External function structure
//

struct NXSL_ExtFunction
{
   char m_szName[MAX_FUNCTION_NAME];
   int (* m_pfHandler)(int argc, NXSL_Value **argv, NXSL_Value **ppResult);
   int m_iNumArgs;   // Number of arguments or -1 fo variable number
};


//
// Environment for NXSL program
//

class NXSL_Program;
class NXSL_Library;

class LIBNXSL_EXPORTABLE NXSL_Environment
{
private:
   DWORD m_dwNumFunctions;
   NXSL_ExtFunction *m_pFunctionList;

   NXSL_Library *m_pLibrary;

   FILE *m_pStdIn;
   FILE *m_pStdOut;

public:
   NXSL_Environment();
   ~NXSL_Environment();

   void SetIO(FILE *pIn, FILE *pOut) { m_pStdIn = pIn; m_pStdOut = pOut; }
   FILE *GetStdIn(void) { return m_pStdIn; }
   FILE *GetStdOut(void) { return m_pStdOut; }

   void SetLibrary(NXSL_Library *pLib) { m_pLibrary = pLib; }

   NXSL_ExtFunction *FindFunction(char *pszName);
   void RegisterFunctionSet(DWORD dwNumFunctions, NXSL_ExtFunction *pList);

   BOOL UseModule(NXSL_Program *pMain, char *pszName);
};


//
// Runtime variable information
//

class LIBNXSL_EXPORTABLE NXSL_Variable
{
protected:
   TCHAR *m_pszName;
   NXSL_Value *m_pValue;

public:
   NXSL_Variable(const TCHAR *pszName);
   NXSL_Variable(const TCHAR *pszName, NXSL_Value *pValue);
   NXSL_Variable(NXSL_Variable *pSrc);
   ~NXSL_Variable();

   const TCHAR *Name(void) { return m_pszName; }
   NXSL_Value *Value(void) { return m_pValue; }
   void Set(NXSL_Value *pValue);
};


//
// Varable system
//

class LIBNXSL_EXPORTABLE NXSL_VariableSystem
{
protected:
   DWORD m_dwNumVariables;
   NXSL_Variable **m_ppVariableList;

public:
   NXSL_VariableSystem(void);
   NXSL_VariableSystem(NXSL_VariableSystem *pSrc);
   ~NXSL_VariableSystem();

   NXSL_Variable *Find(const TCHAR *pszName);
   NXSL_Variable *Create(const TCHAR *pszName, NXSL_Value *pValue = NULL);
};


//
// Single execution instruction
//

class LIBNXSL_EXPORTABLE NXSL_Instruction
{
   friend class NXSL_Program;

protected:
   int m_nOpCode;
   union
   {
      NXSL_Value *m_pConstant;
      char *m_pszString;
      DWORD m_dwAddr;
   } m_operand;
   int m_nStackItems;
   int m_nSourceLine;

public:
   NXSL_Instruction(int nLine, int nOpCode);
   NXSL_Instruction(int nLine, int nOpCode, NXSL_Value *pValue);
   NXSL_Instruction(int nLine, int nOpCode, char *pszString);
   NXSL_Instruction(int nLine, int nOpCode, char *pszString, int nStackItems);
   NXSL_Instruction(int nLine, int nOpCode, DWORD dwAddr);
   NXSL_Instruction(int nLine, int nOpCode, int nStackItems);
   NXSL_Instruction(NXSL_Instruction *pSrc);
   ~NXSL_Instruction();
};


//
// Used module information
//

struct NXSL_Module
{
   char m_szName[MAX_PATH];
   DWORD m_dwCodeStart;
   DWORD m_dwCodeSize;
   DWORD m_dwFunctionStart;
   DWORD m_dwNumFunctions;
};


//
// Class representing compiled NXSL program
//

class LIBNXSL_EXPORTABLE NXSL_Program
{
protected:
   NXSL_Environment *m_pEnv;

   NXSL_Instruction **m_ppInstructionSet;
   DWORD m_dwCodeSize;
   DWORD m_dwCurrPos;

   DWORD m_dwNumPreloads;
   char **m_ppszPreloadList;

   DWORD m_dwSubLevel;
   NXSL_Stack *m_pDataStack;
   NXSL_Stack *m_pCodeStack;
   int m_nBindPos;

   NXSL_VariableSystem *m_pConstants;
   NXSL_VariableSystem *m_pGlobals;
   NXSL_VariableSystem *m_pLocals;

   DWORD m_dwNumFunctions;
   NXSL_Function *m_pFunctionList;

   DWORD m_dwNumModules;
   NXSL_Module *m_pModuleList;

   NXSL_Value *m_pRetValue;
   int m_nErrorCode;
   TCHAR *m_pszErrorText;

   void Execute(void);
   void CallFunction(int nArgCount);
   void DoUnaryOperation(int nOpCode);
   void DoBinaryOperation(int nOpCode);
   void Error(int nError);
   NXSL_Value *MatchRegexp(NXSL_Value *pValue, NXSL_Value *pRegexp, BOOL bIgnoreCase);

   NXSL_Variable *FindOrCreateVariable(TCHAR *pszName);
	NXSL_Variable *CreateVariable(TCHAR *pszName);

   DWORD GetFunctionAddress(char *pszName);
   void RelocateCode(DWORD dwStartOffset, DWORD dwLen, DWORD dwShift);
	DWORD FinalJumpDestination(DWORD dwAddr);

public:
   NXSL_Program(void);
   ~NXSL_Program();

   BOOL AddFunction(const char *pszName, DWORD dwAddr, char *pszError);
   void ResolveFunctions(void);
   void AddInstruction(NXSL_Instruction *pInstruction);
   void ResolveLastJump(int nOpCode);
	void CreateJumpAt(DWORD dwOpAddr, DWORD dwJumpAddr);
   void AddPreload(char *pszName);
   void UseModule(NXSL_Program *pModule, const char *pszName);
	void Optimize(void);

	void SetGlobalVariable(const TCHAR *pszName, NXSL_Value *pValue);
	NXSL_Variable *FindGlobalVariable(const TCHAR *pszName) { return m_pGlobals->Find(pszName); }

   int Run(NXSL_Environment *pEnv = NULL, DWORD argc = 0,
           NXSL_Value **argv = NULL, NXSL_VariableSystem *pUserLocals = NULL,
           NXSL_VariableSystem **ppGlobals = NULL);

   DWORD CodeSize(void) { return m_dwCodeSize; }

   void Dump(FILE *pFile);
   TCHAR *GetErrorText(void) { return m_pszErrorText; }
   NXSL_Value *GetResult(void) { return m_pRetValue; }
};


//
// Script library
//

class LIBNXSL_EXPORTABLE NXSL_Library
{
private:
   DWORD m_dwNumScripts;
   NXSL_Program **m_ppScriptList;
   char **m_ppszNames;
   DWORD *m_pdwIdList;
   MUTEX m_mutex;

   void Delete(int nIndex);

public:
   NXSL_Library(void);
   ~NXSL_Library();

   void Lock(void) { MutexLock(m_mutex, INFINITE); }
   void Unlock(void) { MutexUnlock(m_mutex); }

   BOOL AddScript(DWORD dwId, char *pszName, NXSL_Program *pScript);
   void DeleteScript(char *pszName);
   void DeleteScript(DWORD dwId);
   NXSL_Program *FindScript(char *pszName);

   void FillMessage(CSCPMessage *pMsg);
};


#endif
