%{

#pragma warning(disable : 4065 4102)

#define YYERROR_VERBOSE
#define YYINCLUDED_STDLIB_H
#define YYDEBUG			1

#include <nms_common.h>
#include "libnxsl.h"
#include "parser.tab.hpp"

void yyerror(yyscan_t scanner, NXSL_Lexer *pLexer, NXSL_Compiler *pCompiler,
             NXSL_Program *pScript, const char *pszText);
int yylex(YYSTYPE *lvalp, yyscan_t scanner);

%}

%expect		1
%pure-parser
%lex-param		{yyscan_t scanner}
%parse-param	{yyscan_t scanner}
%parse-param	{NXSL_Lexer *pLexer}
%parse-param	{NXSL_Compiler *pCompiler}
%parse-param	{NXSL_Program *pScript}

%union
{
	LONG valInt32;
	DWORD valUInt32;
	INT64 valInt64;
	QWORD valUInt64;
	char *valStr;
	double valReal;
	NXSL_Value *pConstant;
	NXSL_Instruction *pInstruction;
}

%token T_ARRAY
%token T_BREAK
%token T_CASE
%token T_CONTINUE
%token T_DEFAULT
%token T_DO
%token T_ELSE
%token T_EXIT
%token T_FOR
%token T_IF
%token T_NULL
%token T_PRINT
%token T_PRINTLN
%token T_RETURN
%token T_SUB
%token T_SWITCH
%token T_TYPE_INT32
%token T_TYPE_INT64
%token T_TYPE_REAL
%token T_TYPE_STRING
%token T_TYPE_UINT32
%token T_TYPE_UINT64
%token T_USE
%token T_WHILE

%token <valStr> T_COMPOUND_IDENTIFIER
%token <valStr> T_IDENTIFIER
%token <valStr> T_STRING
%token <valInt32> T_INT32
%token <valUInt32> T_UINT32
%token <valInt64> T_INT64
%token <valUInt64> T_UINT64
%token <valReal> T_REAL

%right '='
%right ':'
%left '?'
%left '.'
%left T_OR
%left T_AND
%left '|'
%left '^'
%left '&'
%left T_NE T_EQ T_LIKE T_ILIKE T_MATCH T_IMATCH
%left '<' T_LE '>' T_GE
%left T_LSHIFT T_RSHIFT
%left '+' '-'
%left '*' '/' '%'
%right T_INC T_DEC '!' '~' NEGATE
%left T_POST_INC T_POST_DEC T_REF '[' ']'

%type <pConstant> Constant
%type <valStr> AnyIdentifier
%type <valStr> FunctionName
%type <valInt32> BuiltinType
%type <valInt32> ParameterList
%type <pInstruction> SimpleStatementKeyword

%start Script


%%

Script:
	Module
|	Expression
{
	char szErrorText[256];

	// Add implicit return
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_RETURN));

	// Add implicit main() function
	if (!pScript->AddFunction("main", 0, szErrorText))
	{
		pCompiler->Error(szErrorText);
		pLexer->SetErrorState();
	}
}
;

Module:
	ModuleComponent Module
|	ModuleComponent
;

ModuleComponent:
	Function
|	UseStatement
;

UseStatement:
	T_USE AnyIdentifier ';'
{
	pScript->AddPreload($2);
}
;

AnyIdentifier:
	T_IDENTIFIER
|	T_COMPOUND_IDENTIFIER
;

Function:
	T_SUB FunctionName 
	{
		char szErrorText[256];

		if (!pScript->AddFunction($2, INVALID_ADDRESS, szErrorText))
		{
			pCompiler->Error(szErrorText);
			pLexer->SetErrorState();
		}
		free($2);
		pCompiler->SetIdentifierOperation(OPCODE_BIND);
	}
	ParameterDeclaration Block
	{
		pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_RET_NULL));
	}
;

ParameterDeclaration:
	IdentifierList ')'
|	')'
;

IdentifierList:
	T_IDENTIFIER 
	{
		pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), pCompiler->GetIdentifierOperation(), $1));
	}
	',' IdentifierList
|	T_IDENTIFIER
	{
		pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), pCompiler->GetIdentifierOperation(), $1));
	}
;

Block:
	'{' StatementList '}'
;

StatementList:
	StatementOrBlock StatementList
|
;

StatementOrBlock:
	Statement
|	Block
;

Statement:
	Expression ';'
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_POP, (int)1));
}
|	BuiltinStatement
|	';'
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_NOP));
}
;

Expression:
	'(' Expression ')'
|	T_IDENTIFIER '=' Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_SET, $1));
}
|	Expression '[' Expression ']' '=' Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_SET_ELEMENT));
}
|	Expression '[' Expression ']'
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_GET_ELEMENT));
}
|	Expression T_REF T_IDENTIFIER
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_REFERENCE, $3));
}
|	'-' Expression		%prec NEGATE
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_NEG));
}
|	'!' Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_NOT));
}
|	'~' Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_BIT_NOT));
}
|	T_INC T_IDENTIFIER
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_INCP, $2));
}
|	T_DEC T_IDENTIFIER
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_DECP, $2));
}
|	T_IDENTIFIER T_INC	%prec T_POST_INC
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_INC, $1));
}
|	T_IDENTIFIER T_DEC	%prec T_POST_DEC
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_DEC, $1));
}
|	Expression '+' Expression	
{ 
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_ADD));
}
|	Expression '-' Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_SUB));
}
|	Expression '*' Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_MUL));
}
|	Expression '/' Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_DIV));
}
|	Expression '%' Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_REM));
}
|	Expression T_LIKE Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_LIKE));
}
|	Expression T_ILIKE Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_ILIKE));
}
|	Expression T_MATCH Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_MATCH));
}
|	Expression T_IMATCH Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_IMATCH));
}
|	Expression T_EQ Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_EQ));
}
|	Expression T_NE Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_NE));
}
|	Expression '<' Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_LT));
}
|	Expression T_LE Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_LE));
}
|	Expression '>' Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_GT));
}
|	Expression T_GE Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_GE));
}
|	Expression '&' Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_BIT_AND));
}
|	Expression '|' Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_BIT_OR));
}
|	Expression '^' Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_BIT_XOR));
}
|	Expression T_AND Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_AND));
}
|	Expression T_OR Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_OR));
}
|	Expression T_LSHIFT Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_LSHIFT));
}
|	Expression T_RSHIFT Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_RSHIFT));
}
|	Expression '.' Expression
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_CONCAT));
}
|	Expression '?'
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_JZ, INVALID_ADDRESS));
}
	Expression ':'
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_JMP, INVALID_ADDRESS));
	pScript->ResolveLastJump(OPCODE_JZ);
}	 
	Expression
{
	pScript->ResolveLastJump(OPCODE_JMP);
}
|	Operand
;

Operand:
	FunctionCall
|	TypeCast
|	T_IDENTIFIER
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_PUSH_VARIABLE, $1));
}
|	Constant
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_PUSH_CONSTANT, $1));
}
;

TypeCast:
	BuiltinType '(' Expression ')'
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(),
				OPCODE_CAST, (int)$1));
}
;

BuiltinType:
	T_TYPE_INT32
{
	$$ = NXSL_DT_INT32;
}
|	T_TYPE_INT64
{
	$$ = NXSL_DT_INT64;
}
|	T_TYPE_UINT32
{
	$$ = NXSL_DT_INT32;
}
|	T_TYPE_UINT64
{
	$$ = NXSL_DT_UINT64;
}
|	T_TYPE_REAL
{
	$$ = NXSL_DT_REAL;
}
|	T_TYPE_STRING
{
	$$ = NXSL_DT_STRING;
}
;

BuiltinStatement:
	SimpleStatement ';'
|	PrintlnStatement
|	IfStatement
|	DoStatement
|	WhileStatement
|	SwitchStatement
|	ArrayStatement
|	T_BREAK ';'
{
	if (pCompiler->CanUseBreak())
	{
		pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_NOP));
		pCompiler->AddBreakAddr(pScript->CodeSize() - 1);
	}
	else
	{
		pCompiler->Error("\"break\" statement can be used only within loops and \"switch\" statements");
		pLexer->SetErrorState();
	}
}
|	T_CONTINUE ';'
{
	DWORD dwAddr = pCompiler->PeekAddr();
	if (dwAddr != INVALID_ADDRESS)
	{
		pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_JMP, dwAddr));
	}
	else
	{
		pCompiler->Error("\"continue\" statement can be used only within loops");
		pLexer->SetErrorState();
	}
}
;

SimpleStatement:
	SimpleStatementKeyword Expression
{
	pScript->AddInstruction($1);
}
|	SimpleStatementKeyword
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_PUSH_CONSTANT, new NXSL_Value));
	pScript->AddInstruction($1);
}
;

SimpleStatementKeyword:
	T_EXIT
{
	$$ = new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_EXIT);
}
|	T_RETURN
{
	$$ = new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_RETURN);
}
|	T_PRINT
{
	$$ = new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_PRINT);
}
;

PrintlnStatement:
	T_PRINTLN Expression ';'
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_PUSH_CONSTANT, new NXSL_Value(_T("\n"))));
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_CONCAT));
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_PRINT));
}
|	T_PRINTLN ';'
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_PUSH_CONSTANT, new NXSL_Value(_T("\n"))));
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_PRINT));
}
;

IfStatement:
	T_IF '(' Expression ')' 
	{
		pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_JZ, INVALID_ADDRESS));
	} 
	IfBody
;

IfBody:
	StatementOrBlock
{
	pScript->ResolveLastJump(OPCODE_JZ);
}
|	StatementOrBlock ElseStatement
{
	pScript->ResolveLastJump(OPCODE_JMP);
}
;

ElseStatement:
	T_ELSE
	{
		pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_JMP, INVALID_ADDRESS));
		pScript->ResolveLastJump(OPCODE_JZ);
	}
	StatementOrBlock
;

WhileStatement:
	T_WHILE
{
	pCompiler->PushAddr(pScript->CodeSize());
	pCompiler->NewBreakLevel();
}
	'(' Expression ')'
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_JZ, INVALID_ADDRESS));
}
	StatementOrBlock
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), 
				OPCODE_JMP, pCompiler->PopAddr()));
	pScript->ResolveLastJump(OPCODE_JZ);
	pCompiler->CloseBreakLevel(pScript);
}
;

DoStatement:
	T_DO
{
	pCompiler->PushAddr(pScript->CodeSize());
	pCompiler->NewBreakLevel();
}
	StatementOrBlock T_WHILE '(' Expression ')' ';'
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(),
				OPCODE_JNZ, pCompiler->PopAddr()));
	pCompiler->CloseBreakLevel(pScript);
}
;

SwitchStatement:
	T_SWITCH
{ 
	pCompiler->NewBreakLevel();
}
	'(' Expression ')' '{' CaseList Default '}'
{
	pCompiler->CloseBreakLevel(pScript);
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_POP, (int)1));
}
;

CaseList:
	Case
{ 
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_JMP, pScript->CodeSize() + 3));
	pScript->ResolveLastJump(OPCODE_JZ);
}
	CaseList
|	Case
{
	pScript->ResolveLastJump(OPCODE_JZ);
}
;

Case:
	T_CASE Constant
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_CASE, $2));
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_JZ, INVALID_ADDRESS));
} 
	':' StatementList
;

Default:
	T_DEFAULT ':' StatementList
|
;

ArrayStatement:
	T_ARRAY { pCompiler->SetIdentifierOperation(OPCODE_ARRAY); } IdentifierList ';'
;

FunctionCall:
	FunctionName ParameterList ')'
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_CALL_EXTERNAL, $1, $2));
}
|	FunctionName ')'
{
	pScript->AddInstruction(new NXSL_Instruction(pLexer->GetCurrLine(), OPCODE_CALL_EXTERNAL, $1, 0));
}
;

ParameterList:
	Expression ',' ParameterList { $$ = $3 + 1; }
|	Expression { $$ = 1; }
;

FunctionName:
	T_IDENTIFIER '('
{
	$$ = $1;
}
;

Constant:
	T_STRING
{
	$$ = new NXSL_Value($1);
	free($1);
}
|	T_INT32
{
	$$ = new NXSL_Value($1);
}
|	T_UINT32
{
	$$ = new NXSL_Value($1);
}
|	T_INT64
{
	$$ = new NXSL_Value($1);
}
|	T_UINT64
{
	$$ = new NXSL_Value($1);
}
|	T_REAL
{
	$$ = new NXSL_Value($1);
}
|	T_NULL
{
	$$ = new NXSL_Value;
}
;
