/* 
** NetXMS - Network Management System
** Copyright (C) 2003, 2004, 2005, 2006 Victor Kirhenshtein
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
** File: debug.cpp
**
**/

#include "nxcore.h"

#ifdef _WIN32
#include <dbghelp.h>
#endif


//
// Test mutex state and print to stdout
//

void DbgTestMutex(MUTEX hMutex, const TCHAR *szName, CONSOLE_CTX pCtx)
{
   ConsolePrintf(pCtx, _T("  %s: "), szName);
   if (MutexLock(hMutex, 100))
   {
      ConsolePrintf(pCtx, _T("unlocked\n"));
      MutexUnlock(hMutex);
   }
   else
   {
      ConsolePrintf(pCtx, _T("locked\n"));
   }
}


//
// Test read/write lock state and print to stdout
//

void DbgTestRWLock(RWLOCK hLock, const TCHAR *szName, CONSOLE_CTX pCtx)
{
   ConsolePrintf(pCtx, _T("  %s: "), szName);
   if (RWLockWriteLock(hLock, 100))
   {
      ConsolePrintf(pCtx, _T("unlocked\n"));
      RWLockUnlock(hLock);
   }
   else
   {
      if (RWLockReadLock(hLock, 100))
      {
         ConsolePrintf(pCtx, _T("locked for reading\n"));
         RWLockUnlock(hLock);
      }
      else
      {
         ConsolePrintf(pCtx, _T("locked for writing\n"));
      }
   }
}


//
// Print message to console, either local or remote
//

void ConsolePrintf(CONSOLE_CTX pCtx, const char *pszFormat, ...)
{
   va_list args;

   va_start(args, pszFormat);
   if (pCtx->hSocket == -1)
   {
      vprintf(pszFormat, args);
   }
   else
   {
      CSCP_MESSAGE *pRawMsg;
      TCHAR szBuffer[8192];

      _vsntprintf(szBuffer, 8191, pszFormat, args);
		szBuffer[8191] = 0;
      pCtx->pMsg->SetVariable(VID_MESSAGE, szBuffer);
      pRawMsg = pCtx->pMsg->CreateMessage();
      SendEx(pCtx->hSocket, pRawMsg, ntohl(pRawMsg->dwSize), 0);
      free(pRawMsg);
   }
   va_end(args);
}


//
// Show server statistics
//

void ShowServerStats(CONSOLE_CTX pCtx)
{
   DWORD i, dwNumItems;

   RWLockReadLock(g_rwlockNodeIndex, INFINITE);
   for(i = 0, dwNumItems = 0; i < g_dwNodeAddrIndexSize; i++)
      dwNumItems += ((Node *)g_pNodeIndexByAddr[i].pObject)->GetItemCount();
   RWLockUnlock(g_rwlockNodeIndex);

   ConsolePrintf(pCtx, "Total number of objects:     %d\n"
                       "Number of monitored nodes:   %d\n"
                       "Number of collectable DCIs:  %d\n\n",
                 g_dwIdIndexSize, g_dwNodeAddrIndexSize, dwNumItems);
}


//
// Show queue stats
//

void ShowQueueStats(CONSOLE_CTX pCtx, Queue *pQueue, const char *pszName)
{
   if (pQueue != NULL)
      ConsolePrintf(pCtx, "%-32s : %d\n", pszName, pQueue->Size());
}


//
// Write process coredump
//

#ifdef _WIN32

void DumpProcess(CONSOLE_CTX pCtx)
{
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	char cmdLine[64], buffer[256];

	ConsolePrintf(pCtx, "Dumping process to disk...\n");

	sprintf(cmdLine, "netxmsd.exe --dump %d", GetCurrentProcessId());
	memset(&si, 0, sizeof(STARTUPINFO));
	si.cb = sizeof(STARTUPINFO);
	if (CreateProcess(NULL, cmdLine, NULL, NULL, FALSE,
	                  (g_dwFlags & AF_DAEMON) ? CREATE_NO_WINDOW : 0, NULL, NULL, &si, &pi))
	{
		WaitForSingleObject(pi.hProcess, INFINITE);
		CloseHandle(pi.hThread);
		CloseHandle(pi.hProcess);
		
		ConsolePrintf(pCtx, "Done.\n");
	}
	else
	{
		ConsolePrintf(pCtx, "Dump error: CreateProcess() failed (%s)\n", GetSystemErrorText(GetLastError(), buffer, 256));
	}
}

#else

void DumpProcess(CONSOLE_CTX pCtx)
{
	ConsolePrintf(pCtx, "DUMP command is not supported for current operating system\n");
}

#endif
